// ---------------------------------------------------------------------------
// Random Image Viewer
//
// Copyright IBM Corporation 1997
// See HTTP://www.ibm.com/java/apps/license.html for license information.
//
// Randomly display an image from a list of images each time the page
// containing the applet is displayed.  If the user clicks on the applet,
// the browser is directed to go to the URL associated with the image.
//
// The following shows an example coding of the applet tag for the applet.
// imagen and URLn are each considered to be a pair.  The image and URL
// params start at 1 and go up sequentially.  The first missing param
// in the sequence signals the end of the params.
// <applet code="RandomImageViewer.class" codebase="." width=100 height=100>
//  <param name=image1 value="image1.jpg">
//  <param name=URL1   value="http://wwwidd.ibm.com/~jgraham/searchRec.html">
//  <param name=image2 value="image2.jpg">
//  <param name=URL2   value="http://wwwidd.ibm.com/~jgraham/welcome.html">
//  <param name=image3 value="image3.jpg">
//  <param name=URL3   value="http://www.ibm.com/Java">
// </applet>
//
// Note it is possible that the same image may be displayed consecutively, if
// the random number generator returns the same number twice in a row.  You
// decrease the chance of this happening by having a larger list of images and
// URLs.
// 
// Date    Int  Action
// ------- ---  ------------------------------------
// 7/18/97 jsg  Created
// ---------------------------------------------------------------------------

import java.awt.*;
import java.util.*;
import java.lang.*;
import java.net.*;

public class RandomImageViewer extends java.applet.Applet  implements Runnable {

  private MediaTracker mt;
  private Random randGen;

  // image and URL params start at 1 and go up sequentially.  All arrays
  // based on the params start with index 0, so the index of the array element
  // derived from a param is 1 less than the number on the corresponding param.
  private final int paramLow=1;  // first image param is image1 & URL is URL1
  private int paramHigh;         // last image & URL param
  private int curIndex;          // index of current image & url (param#-1)
  private Image[] images;        // images we have loadedfrom image params
  private URL[] urls;            // URLs built from URL params
  private String[] urlStrings;   // string form of URLs from URL params
  private String[] imageStrings; // image names from image params

  // ---------------------------------------------------------------------------
  // When applet is started, load the params and start loading the first image.
  // ---------------------------------------------------------------------------
   public void init() {
      mt=new MediaTracker(this);
      randGen=new Random();

      //  Load image and URL params.
      String[] pa={"image","URL"};          // prefixes of params
      String[][] a=paramArray(pa,paramLow);
      paramHigh=a[0].length;
      imageStrings=a[0];
      urlStrings=a[1];
      images=new Image[paramHigh];

      //  Start load of first image
      curIndex=randRange(randGen,paramLow,paramHigh)-1;  // random new curIndex
      images[curIndex]=getImage(getDocumentBase(),imageStrings[curIndex]);
      mt.addImage(images[curIndex],curIndex);
      Thread t=new Thread(this);
      t.setPriority(Thread.MAX_PRIORITY);  // get run to go before applet start
      t.start();

      // Create URLs from URL params
      urls=new URL[paramHigh];
      for (int i=0; i<paramHigh; i++) {
	try {
	  urls[i]=new URL(urlStrings[i]);
	}  
	catch(java.net.MalformedURLException e) {
	  System.err.println("Syntactically incorrect URL: "+
			     urlStrings[i]);
	  System.exit(1);
	}
      }
      
   }  // end init

  // ---------------------------------------------------------------------------
  // Display currently loaded image
  // ---------------------------------------------------------------------------
  public void paint(Graphics g) {
    g.drawImage(images[curIndex],0,0,this);
  } // end paint

  // ---------------------------------------------------------------------------
  // Show current destination URL in status area when mouse over applet
  // ---------------------------------------------------------------------------
  public boolean mouseEnter(Event e, int x, int y) {
    showStatus(urlStrings[curIndex]);
    return true;
  } // end mouseEnter

  // ---------------------------------------------------------------------------
  // Blank status area when mouse leaves applet
  // ---------------------------------------------------------------------------
  public boolean mouseExit(Event e, int x, int y) {
    showStatus("");
    return true;
  } // end mouseExit

  // ---------------------------------------------------------------------------
  // Tell browser to go to destination URL when user clicks on applet
  // ---------------------------------------------------------------------------
  public boolean mouseDown(Event e, int x, int y) {
    getAppletContext().showDocument(urls[curIndex]);
    return true;
  }  // end mouseDown

  // ---------------------------------------------------------------------------
  // When browser leaves page with applet, start load of next image.
  // ---------------------------------------------------------------------------
  public void stop() {
    curIndex=randRange(randGen,paramLow,paramHigh)-1; // random new curIndex
    if (images[curIndex] == null) {
      images[curIndex]=getImage(getDocumentBase(),imageStrings[curIndex]);
      mt.addImage(images[curIndex],curIndex);
      (new Thread(this)).start();
    }
  }  // end stop

  // ---------------------------------------------------------------------------
  // thread that issues the wait for the image to start the load
  // ---------------------------------------------------------------------------
  public void run() {
    try {mt.waitForID(curIndex);}
    catch (Exception e) {
      System.err.println("Error occured in run while loading "+
			 imageStrings[curIndex]);
      e.printStackTrace();
    }
    if (mt.isErrorID(curIndex)) {
      System.err.println("ErrorID in run reported error occured while loading "+
			 imageStrings[curIndex]);
    }
  }  // end run

  // ---------------------------------------------------------------------------
  // When applet is first started or when user comes back to page with applet,
  // wait for image to finish loading.  paint() will display the image.
  // ---------------------------------------------------------------------------
  public void start() {
    try {mt.waitForID(curIndex);}
    catch (Exception e) {
      System.err.println("Error occured while loading "+
			 imageStrings[curIndex]);
      e.printStackTrace();
    }
    if (mt.isErrorID(curIndex)) {
      System.err.println("ErrorID reported error occured while loading "+
			 imageStrings[curIndex]);
    }
  }  // end start

  // ---------------------------------------------------------------------------
  // return random integer within range passed by user
  // ---------------------------------------------------------------------------
  private int randRange(Random randGen,  // instance of Random to use
                        int low,         // lowest number to return
                        int high         // highest number to return
                        ) {
    float fRand=randGen.nextFloat();
    float rangeLength=high - low + 1;
    float randRange=rangeLength * fRand;
    int iRand=(int) randRange + low;
    if (iRand > high) {
      iRand=high;
    }
    return iRand;
  }  // end randRange

  // ---------------------------------------------------------------------------
  // For each element of passed array paramPrefixes, return an array in the
  // 2 dimension array that contains the values of each param starting with
  // the prefix.  For example, if paramPrefixes={"image","URL"}, paramLow=1
  // and the params on the applet tag are image1=a, image2=b, url1=c, and
  // url2=d, the returned array will be as follows: {{a,b},{c,d}}
  //
  // The params specified by paramPrefixes are considered to start with
  // a suffix of paramLow and to increase sequentially.  The first missing suffix
  // signals the end of the params for that prefix.
  //
  // All of the params specified by paramPrefixes are considered to be a matched
  // set, so it is an error if for a given suffix, params exist for some of the
  // prefixes but not for the others.  When this error occurs, a message is
  // printed on stderr, but processing continues.
  //
  // Note that all the arrays used here have 0 as their first index, regardless
  // of what the value of paramLow is.
  // ---------------------------------------------------------------------------
  private String[][] paramArray(String[] paramPrefixes,  //prefixes of params
                                int paramLow             //first suffix
                                ) {
    // each vector in v will hold values for a prefix
    Vector[] v=new Vector[paramPrefixes.length];
    for (int i=0;i<paramPrefixes.length;i++) {v[i]=new Vector();}
    String[] paramValues=new String[paramPrefixes.length];

    scanparams: for (int i=paramLow; ;i++) {  // suffix we are trying
      for (int j=0; j<paramPrefixes.length; j++) {  // try each prefix for suffix
	paramValues[j]=getParameter(paramPrefixes[j]+i);
      }
      // see if 1 prefix, all prefixes, or no prefixes got a value for this suffix
      boolean got1Parm=false;
      boolean gotAllParm=true;
      for (int j=0; j<paramPrefixes.length; j++) {
	if (paramValues[j] !=null) {
	  got1Parm=true;
	}
	else {
	  gotAllParm=false;
	}
      }
      if (got1Parm)           // 1 prefix got a value
	if (gotAllParm)       // all prefixes got a value
          // add each value to vector holding values for the prefix
	  for (int j=0; j<paramPrefixes.length; j++) {
	    String x=paramValues[j];
	    v[j].addElement(x);
	  }
	else {               // only some prefixes got a value
          //  warn user which prefixes were missing a value
	  for (int j=0; j<paramPrefixes.length; j++) {
	    if (paramValues[j] == null)
	      System.err.println("Missing param " + paramPrefixes[j]+i +"\n");
	  }
	  break scanparams;  // stop checking for params
	}
      else {                 // no prefixes had a value...this is the end
    	break scanparams;    // stop checking for params
      }
    }

    // copy values for each prefix into the 2 dim array we return
    String[][]params = new String[paramPrefixes.length][v[0].size()];
    for (int j=0; j<paramPrefixes.length; j++) {
      v[j].copyInto(params[j]);
    }
    return params;
  }  //  end of paramArray
  
}  // end of class RandomImageViewer
