/******************************************************************************/
/* COPYRIGHT: Copyright (C) International Business Machines Corp., 1997       */
/*                                                                            */
/* DISCLAIMER OF WARRANTIES:                                                  */
/*   The following [enclosed] code is sample code created by IBM              */
/*   Corporation.  This sample code is not part of any standard IBM product   */
/*   and is provided to you solely for the purpose of assisting you in the    */
/*   development of your applications.  The code is provided "AS IS",         */
/*   without warranty of any kind.  IBM shall not be liable for any damages   */
/*   arising out of your use of the sample code, even if they have been       */
/*   advised of the possibility of such damages.                              */
/******************************************************************************/

import java.lang.ThreadGroup;
import java.lang.Throwable;

/*
 * Class: Vendor
 *
 * This class is written by a fictitious vendor.
 * For our scenarios, it is a black box -- we have no
 * access to the source; in fact we only purchase a .CLASS file.
 *
 * Unfortunately, the class was developed by a hacker, and not
 * well tested. It eventually throws a NullPointerException,
 * which is a subclass of RunTimeException -- one of those
 * nasty exceptions that doesn't have to be advertised in a throws
 * clause.
 *
 */

class Vendor
{
  public void doIt() 
  {
    for (int i=0; i<10; i++) 
    {
      System.out.println("Vendor thread: [" + Thread.currentThread().getName() +"]");
      if (i==9) 
         throw new NullPointerException("Null Pointer Exception thrown from Vendor thread");
    }
  }
}

/*
 * Class: CatchableThreadException
 *
 * This class is used by a thread wrapper to encapsulate some additional
 * information in the throw to the calling code. If the thread wrapper simply
 * rethrew the exception that it obtained from its uncaughtException() method,
 * the identity of the offending secondary thread would be lost, making 
 * intelligent handling of multiple secondary threads impossible.
 *
 * Two getters are provided to make the additional information available.
 *
 * Being a RuntimeException gives this class the freedom to potentially be thrown
 * from any method, without having to advertise in a throws clause
 * (for example, in the run() method of a Runnable).
 * 
 */
class CatchableThreadException extends RuntimeException
{
   private
      Thread _sourceThread;         // Thread in which exception occurred
      Throwable _threadException;   // Actual exception that was caught

   public CatchableThreadException(Thread sourceThread,
                                   Throwable threadException)
   {
      super("Exception caught in thread ["+sourceThread.getName()+"]");
      System.out.println("CatchableThreadException::CTOR");
      
      _sourceThread    = sourceThread;
      _threadException = threadException;
   }

   public Thread getSourceThread() {
      return _sourceThread;
   }
   
   public Throwable getThreadException() {
      return _threadException;
   }
}


/*
 * Class: ThreadWrapper
 *
 * This class "wrappers" the thread class to allow catchable exceptions from
 * within its run() implementation. The wrapper actually is a subclass
 * of (is-a) ThreadGroup, since we have to override uncaughtException.
 * To achieve thread functionality, the wrapper aggregates (has-a)
 * Thread.
 *
 * This wrapper version is designed to handle a single thread only,
 * and actually blocks on thread completion using join(). This allows us
 * to properly catch and rethrow the exception for subsequent catching
 * by the calling code. If we did not block, the main thread would
 * keep on executing after start(), falling through the try
 * block and causing it to lose the rethrown exception.
 * 
 */
 
class ThreadWrapper extends ThreadGroup 
{
   private
      Thread _thread;
      Thread _sourceThread;        // Thread in which exception occurred
      Throwable _threadException;  // Exception caught by uncaughtException()
      boolean exceptionThrown = false;
   
   public ThreadWrapper(Runnable target, String name)
   { 
      super("ThreadWrapper thread group");
      System.out.println("ThreadWrapper::CTOR");
      
      // Create this thread as a member in our thread group
      _thread = new Thread(this,target,name);
   }
   
   public void start() throws CatchableThreadException
   {
      System.out.println("+ThreadWrapper::start()");
      
      _thread.start();
      
      try { _thread.join(); } catch (java.lang.InterruptedException IE) {}
      
      if (exceptionThrown) {
         throw new CatchableThreadException(_sourceThread,_threadException);
      }
      
      System.out.println("-ThreadWrapper::start()");
   }

   public void uncaughtException(Thread sourceThread, Throwable threadException)
   {
      System.out.println("ThreadWrapper::uncaughtException!!!");
      System.out.println(" --> originated in Thread: [" +
                         //Thread.currentThread().getName() + "]");
                         sourceThread.getName() + "]");
   
      // On return from the join() in start(), we want to rethrow an 
      // exception back to the calling code for appropriate handling.
      // Save pertinent information, and set a boolean to inform start().
      _sourceThread    = sourceThread;
      _threadException = threadException;
      exceptionThrown = true;
   }
}

/**
 * Class: threadex2
 *
 * This class demonstrates how an exception can be caught from a 
 * secondary thread using a try-catch block around the start() method 
 * of the ThreadWrapper class.
 *
 * The call to start() actually blocks on the completion of the thread
 * (essentially a synchronous call). If an exception occurs in the thread,
 * a CatchableThreadException is thrown by start() for handling by the
 * calling code.
 *
 */

public class threadex2 implements Runnable 
{
   public static void main(String args[]) {
      threadex2 vendorWrap = new threadex2();   // threadex2 is a Runnable that is a
                                                // client of Vendor

      ThreadWrapper t = new ThreadWrapper(vendorWrap,"Test thread #1");
      
      try {
         System.out.println("Before call to start()");
         t.start();
         System.out.println("After call to start()");
      }
      catch (CatchableThreadException CTE) {
         System.out.println("Details:");
         System.out.println("  Thread name: ["+CTE.getSourceThread().getName()+"]");
         System.out.println("  Exception:   ["+CTE.getThreadException()+"]");

      // Perform other processing or rethrow as appropriate
      // ...
      // throw CTE; 
      }
   }
   
   public void run(){
      // Execute the vendor's implementation in a thread
      Vendor v = new Vendor();
      v.doIt();
   }
}

