/******************************************************************************/
/* COPYRIGHT: Copyright (C) International Business Machines Corp., 1997       */
/*                                                                            */
/* DISCLAIMER OF WARRANTIES:                                                  */
/*   The following [enclosed] code is sample code created by IBM              */
/*   Corporation.  This sample code is not part of any standard IBM product   */
/*   and is provided to you solely for the purpose of assisting you in the    */
/*   development of your applications.  The code is provided "AS IS",         */
/*   without warranty of any kind.  IBM shall not be liable for any damages   */
/*   arising out of your use of the sample code, even if they have been       */
/*   advised of the possibility of such damages.                              */
/******************************************************************************/

import java.awt.*;
import java.awt.event.*;
import java.lang.ThreadGroup;
import java.lang.Throwable;

/*
 * Class: ThreadExceptionCleanup
 *
 * This interface defines the methods that a class should implement
 * to cleanup for a thread after it has thrown an exception and before it is shut down
 * by the system.
 *
 */
interface ThreadExceptionCleanup
{
  public void cleanupOnException(Throwable threadException);
}

/*
 * Class: ThreadExceptionListener
 *
 * This interface defines the methods that a class should implement
 * to handle exceptions that are thrown in its secondary threads.
 *
 */
interface ThreadExceptionListener
{
  public void exceptionOccurred(Thread sourceThread,
                                Throwable threadException);
}

/*
 * Class: Vendor
 *
 * This class is written by a fictitious vendor.
 * For our scenarios, it is a black box -- we have no
 * access to the source; in fact we only purchase a .CLASS file.
 *
 * Unfortunately, the class was developed by a hacker, and not
 * well tested. It eventually throws a NullPointerException,
 * which is a subclass of RunTimeException -- one of those
 * exceptions that doesn't have to be advertised in a throws clause.
 *
 */

class Vendor
{
  public void doIt() 
  {
    for (int i=0; i<10; i++) 
    {
      System.out.println("Vendor thread: [" + Thread.currentThread().getName() +"]");
      if (i==9) 
         throw new NullPointerException("Null Pointer Exception thrown from Vendor thread");
      try { Thread.sleep(300); } catch (Exception e) {}
    }
  }
}

/*
 * Class: ThreadWrapper2
 *
 * This class "wrappers" the thread class to allow catchable exceptions from
 * within its run() implementation. The wrapper actually is a subclass
 * of (is-a) ThreadGroup, since we have to override uncaughtException.
 * To achieve thread functionality, the wrapper aggregates (has-a)
 * Thread.
 *
 */

class ThreadWrapper2 extends ThreadGroup 
{
   private
      ThreadExceptionListener _exceptionListener; 
      Runnable _runnable;
      Thread _thread;
      boolean exceptionThrown = false;

   public ThreadWrapper2(Runnable runnable, String name)
   { 
      super("ThreadWrapper2 thread group");
      System.out.println("ThreadWrapper2::CTOR");
      
      _runnable = runnable;  
      // Create this thread as a member in our thread group
      _thread = new Thread(this, runnable, name);  
   }
   
  public void addThreadExceptionListener(ThreadExceptionListener exceptionListener)
  {
    _exceptionListener = exceptionListener;
  }
  
  public Runnable getRunnable()
  {
    return _runnable;
  }


   public void start()
   {
      System.out.println("+ThreadWrapper2::start()");
      _thread.start();
      System.out.println("-ThreadWrapper2::start()");
   }

   public void uncaughtException(Thread sourceThread, Throwable threadException)
   {
      System.out.println("ThreadWrapper2::uncaughtException!!!");
      System.out.println(" --> originated in Thread: [" + sourceThread.getName() + "]");
      
      // This thread is running totally asynchronously with the
      // main thread, which is a ThreadExceptionListener,
      // so we just call its exceptionOccurred() method.
      _exceptionListener.exceptionOccurred(sourceThread,threadException); 
      System.out.println("uncaught exiting"); 
   }
}


//This represents some vendor's code that you want to run in a separate thread.  You may or may not have access
//to the source code for this Vendor's product.
class runnableObject implements Runnable, ThreadExceptionCleanup
{
  public void run()
  {
    // Execute the vendor's implementation in a thread
    Vendor v = new Vendor();
    v.doIt();
  } 
  
  public void cleanupOnException(Throwable threadException)
  {
    System.out.println("+CleanupOnException called.");
    System.out.println("Exception received = ["+threadException+"]");
  }
}

class exceptionListener implements ThreadExceptionListener
{
  public void exceptionOccurred(Thread sourceThread, Throwable threadException)
  {
    System.out.println("An exception has occurred in a secondary thread");
    System.out.println("Details:");
    System.out.println("  Thread name: ["+sourceThread.getName()+"]");
    System.out.println("  Exception:   ["+threadException+"]");

    ThreadWrapper2 threadGroup = (ThreadWrapper2)(sourceThread.getThreadGroup());
    runnableObject ro = (runnableObject)(threadGroup.getRunnable());
    ro.cleanupOnException(threadException);
  }
}

class someFrame extends Frame 
{
  public someFrame(String title)
  {
    setTitle(title);
   
    class someFrameAdapter extends WindowAdapter
    {
      public void windowClosing(WindowEvent event) 
      {
        System.exit(0);
      }
    }
    addWindowListener(new someFrameAdapter());
  }
}


/*
 * Class: threadex3
 *
 * This class demonstrates how a ThreadExceptionListener can
 * asynchronously handle exceptions thrown in its secondary threads.
 * Note that try-catch blocks are not required, so the main thread
 * does not have to block on any of its secondaries.
 *
 */
class threadex3
{   
  public static void main(String args[])
  {
    runnableObject runnable = new runnableObject();
    ThreadWrapper2 t = new ThreadWrapper2(runnable, "Test thread #1");
    t.addThreadExceptionListener(new exceptionListener());
    t.start();
    someFrame f = new someFrame("My frame");
    f.setSize(200,200);
    f.show();
  }  
}


